import itertools

import logging
import operator
import time
import re
import json
from functools import reduce

import cm_client
from cm_client import ApiService, ApiRoleConfigGroupList, ApiServiceConfig, ApiClusterTemplateService, \
    ApiClusterTemplateConfig, ApiCommand, ApiParcelList, ApiParcel, ApiConfigList, ApiServiceList, ApiRoleConfigGroup, \
    ApiConfig, ApiRoleNameList, ApiCommandList
from cm_client.rest import ApiException
import requests

LOG = logging.getLogger(__name__)


class CDPCluster:
    """
    Class to represent a CDP Cluster.
    """

    def __init__(self, hostname: str, port: int, is_https: bool, verify_ssl: bool,
                 username: str, password: str, cluster_name: str):
        self.hostname = hostname
        self.port = port
        self.protocol = 'https' if is_https else 'http'
        self.is_https = is_https
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        self.cluster_name = cluster_name

        self._set_configuration()

    def _set_configuration(self):
        LOG.debug("Setting Configuration object")
        configuration = cm_client.Configuration()
        api_version = self._determine_api_version()
        configuration.host = "{protocol}://{hostname}:{port}/api/{api_version}".format(protocol=self.protocol,
                                                                                       hostname=self.hostname,
                                                                                       port=self.port,
                                                                                       api_version=api_version)
        configuration.username = self.username
        configuration.password = self.password
        configuration.verify_ssl = self.verify_ssl

    def _determine_api_version(self):
        response = requests.get(url='{protocol}://{hostname}:{port}/api/version'.format(
            protocol=self.protocol, hostname=self.hostname, port=self.port),
            auth=(self.username, self.password),
            verify=self.verify_ssl)
        if not response.ok:
            LOG.error("Determining API version failed...")
            response.raise_for_status()

        LOG.info("API version: %s" % response.text.strip())
        return response.text.strip()

    def get_cm_version(self, fallback="0"):
        LOG.info("Trying to get the CM version, fallback value: {}".format(fallback))
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.get_version()
            LOG.debug("Response: {}".format(api_response))
            return api_response.version
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->get_version: %s\n" % e)
            return fallback

    def run_service_command(self, service_name: str, command_name: str):
        LOG.info("Running service command '{command}' for service '{service}'...".format(command=command_name,
                                                                                         service=service_name))
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response: ApiCommand = api_instance.service_command_by_name(self.cluster_name, command_name,
                                                                            service_name)
            LOG.debug("Response: %s" % api_response)
            return self.wait_for_service_commands_to_run(service_name, command_name)
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->service_command_by_name: %s\n" % e)
            return False, str(e)

    def get_parcel_info(self, parcel_display_name: str) -> ApiParcel:
        LOG.debug(f"Trying to get details of '{parcel_display_name}'")
        api_instance = cm_client.ParcelsResourceApi(cm_client.ApiClient())

        try:
            api_response: ApiParcelList = api_instance.read_parcels(self.cluster_name, view='full')
            for parcel in api_response.items:
                if parcel.display_name == parcel_display_name:
                    return parcel

            return None
        except ApiException as e:
            LOG.error("Exception when calling ParcelsResourceApi->read_parcels: %s\n" % e)
            raise e

    def download_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 5000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.display_name)
        if current_info.stage != 'AVAILABLE_REMOTELY':
            LOG.info("Parcel is already downloaded, skipping...")
            return True, current_info

        try:
            api_instance.start_download_command(self.cluster_name, parcel_info.product, parcel_info.version)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Starting download of '{parcel_info.display_name}' polling...")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.display_name).stage
                LOG.debug(f"Current stage of '{parcel_info.display_name}': '{current_stage}'")
                if current_stage == 'DOWNLOADED':
                    LOG.info("Download successful...")
                    success = True
                else:
                    time.sleep(30)

            return success, self.get_parcel_info(parcel_info.display_name)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->start_download_command: %s\n" % e)
            raise e

    def distribute_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 5000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.display_name)
        if current_info.stage != 'DOWNLOADED':
            LOG.info("Parcel is already distributed, skipping...")
            return True, current_info

        try:
            result: ApiCommand = api_instance.start_distribution_command(
                self.cluster_name, parcel_info.product, parcel_info.version)

            LOG.info(result)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Starting distribution of '{parcel_info.display_name}' polling...")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.display_name).stage
                LOG.info(self.get_parcel_info(parcel_info.display_name))
                LOG.info(f"Current stage of '{parcel_info.display_name}': '{current_stage}'")
                if current_stage == 'DISTRIBUTED':
                    LOG.info("Distribution successful...")
                    success = True
                else:
                    time.sleep(30)

            return success, self.get_parcel_info(parcel_info.display_name)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->start_distribution_command: %s\n" % e)
            raise e

    def activate_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 1000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.display_name)
        if current_info.stage != 'DISTRIBUTED':
            LOG.info("Parcel is already activated, skipping...")
            return True, current_info

        try:
            api_instance.activate_command(self.cluster_name, parcel_info.product, parcel_info.version)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Activating parcel '{parcel_info.display_name}'.")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.display_name).stage
                LOG.info(f"Current stage of '{parcel_info.display_name}': '{current_stage}'")
                if current_stage == 'ACTIVATED':
                    LOG.info("Activation successful...")
                    success = True
                else:
                    time.sleep(10)

            return success, self.get_parcel_info(parcel_info.display_name)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->activate_parcel: %s\n" % e)
            raise e

    def undistribute_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 5000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.display_name)
        if current_info.stage != 'DISTRIBUTED':
            LOG.info("Parcel is not in distributed state, skipping...")
            return True, current_info

        try:
            api_instance.start_removal_of_distribution_command(
                self.cluster_name, parcel_info.product, parcel_info.version)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Starting undistribution of '{parcel_info.display_name}'. Polling...")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.display_name).stage
                LOG.info(self.get_parcel_info(parcel_info.display_name))
                LOG.info(f"Current stage of '{parcel_info.display_name}': '{current_stage}'")
                if current_stage == 'DOWNLOADED':
                    LOG.info("Undistribution successful...")
                    success = True
                else:
                    time.sleep(15)

            return success, self.get_parcel_info(parcel_info.display_name)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->start_removal_of_distribution_command: %s\n" % e)
            raise e

    def deactivate_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 5000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.display_name)
        if current_info.stage != 'ACTIVATED':
            LOG.info("Parcel is not in activated state, skipping...")
            return True, current_info

        try:
            api_instance.deactivate_command(self.cluster_name, parcel_info.product, parcel_info.version)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Deactivating parcel '{parcel_info.display_name}'.")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.display_name).stage
                LOG.info(f"Current stage of '{parcel_info.display_name}': '{current_stage}'")
                if current_stage == 'DISTRIBUTED':
                    LOG.info("Deactivation successful...")
                    success = True
                else:
                    time.sleep(10)

            return success, self.get_parcel_info(parcel_info.display_name)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->deactivate_command: %s\n" % e)
            raise e

    def deploy_service_client_configuration(self, service_name: str):
        LOG.info("Deploying service client configuration for service '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response: ApiCommand = api_instance.deploy_client_config_command(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)
            return True, api_response.result_message
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->deploy_client_config_command: %s\n" % e)

    def start_service(self, service_name: str, timeout_sec: int = 360):
        LOG.info("Starting service '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.start_command(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)

            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            return self.wait_for_service_state(service_name, expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->start_command: %s\n" % e)
            raise e

    def restart_service(self, service_name: str, timeout_sec: int = 360):
        LOG.info("Restarting service '%s'" % service_name)
        if service_name.lower() == 'mgmt':
            return self.restart_cm_management_service();
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.restart_command(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)

            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            return self.wait_for_service_state(service_name, expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->restart_command: %s\n" % e)
            return False, str(e)

    def read_service(self, service_name: str):
        LOG.debug("Reading service '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.read_service(self.cluster_name, service_name)
            LOG.debug("Response: %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->read_service: %s\n" % e)

    def list_active_role_commands(self, role_name: str, service_name: str):
        LOG.debug("Listing active commands for role '%s' in service '%s'" % (role_name, service_name))
        api_instance = cm_client.RolesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.list_active_commands(self.cluster_name, role_name, service_name, view='summary')
            LOG.debug("Response: %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling RolesResourceApi->list_active_commands: %s\n" % e)

    def list_active_service_commands(self, service_name: str):
        LOG.debug("Listing active commands for service '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.list_active_commands(self.cluster_name, service_name, view='summary')
            LOG.debug("Response: %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling RolesResourceApi->list_active_commands: %s\n" % e)

    def wait_for_service_state(self, service_name: str, expected_states: dict, timeout_sec: int = 360):
        """
        Waits until given service reaches the desired state defined by a dictionary or times out.
        See cm_client.models.api_service.ApiService implementation for attributes that can be checked.

        Example call - this will wait until the service_state is 'STARTED'
                        AND the health_summary is anything other than DISABLED:

            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            self.wait_for_service_state(service_name, expected_service_state, timeout_sec)

        :param service_name: The name of the service
        :param expected_states: Dictionary, where key is the attribute, value is the desired value.
            If value starts with '!', it means that we are expecting the given attribute to NOT equal to the given value
        :param timeout_sec: Maximum timeout in seconds
        :return: (True , message) if service reached the desired state before timeout (False, message) otherwise
        """
        success = False
        message = f"Service {service_name} did not reach desired state in {timeout_sec}s"
        timeout = time.time() + timeout_sec
        while time.time() < timeout and not success:
            if service_name.lower() == 'mgmt':
                service_details: ApiService = self.read_cm_management_service()
            else:
                service_details: ApiService = self.read_service(service_name)
            all_succeeded = True
            for prop, value in expected_states.items():
                invert_logic = value.startswith('!')

                if not invert_logic:
                    if getattr(service_details, prop) != value:
                        all_succeeded = False
                        break
                else:
                    if getattr(service_details, prop) == value[1:]:
                        all_succeeded = False
                        break

            if all_succeeded:
                elapsed_seconds = timeout - time.time()
                success = True
                message = f"Service {service_name} successfully reached desired state in {elapsed_seconds:.2f} seconds"
            else:
                time.sleep(10)

        return success, message

    def wait_for_role_commands_to_run(self, role_name: str, service_name: str, timeout_sec: int = 360):
        """
        Waits until all commands have been executed for a given role of a given service

        :param role_name: The exact name of the role (this should be queried beforehand from read_roles)
        :param service_name: The name of the service
        :param timeout_sec: Maximum timeout in seconds
        :return: (True , message) if all commands have succeeded before timeout (False, message) otherwise
        """
        success = False
        message = f"Commands for {role_name} in service {service_name} did finish in {timeout_sec}s"
        timeout = time.time() + timeout_sec
        while time.time() < timeout and not success:
            active_commands: ApiCommandList = self.list_active_role_commands(role_name, service_name)
            # if the list of currently active commands is empty, we're good to exit
            if not active_commands.items:
                elapsed_seconds = timeout - time.time()
                success = True
                message = f"All commands for {role_name} in service {service_name} have successfully finished in {elapsed_seconds:.2f} seconds"
            else:
                time.sleep(10)

        return success, message

    def wait_for_service_commands_to_run(self, service_name: str, command_name: str, timeout_sec: int = 360):
        """
        Waits until all commands have been executed for a given service

        :param service_name: The name of the service
        :param timeout_sec: Maximum timeout in seconds
        :return: (True , message) if all commands have succeeded before timeout (False, message) otherwise
        """
        success = False
        message = f"Commands for service {service_name} did finish in {timeout_sec}s"
        timeout = time.time() + timeout_sec
        while time.time() < timeout and not success:
            active_commands: ApiCommandList = self.list_active_service_commands(service_name)
            # if the list of currently active commands is empty,
            # or it doesn't contain the command that we're waiting for, we're good to exit
            if not active_commands.items or not list(command
                                                     for command in active_commands.items
                                                     if command.name == command_name):
                elapsed_seconds = timeout - time.time()
                success = True
                message = f"All commands for service {service_name} have successfully finished in {elapsed_seconds:.2f} seconds"
            else:
                time.sleep(10)
        return success, message

    def update_service_config(self, service_name: str, config_dict: dict):
        LOG.info("Updating configs for '%s' - '%s'" % (service_name, config_dict))
        try:
            api_configs_list = []
            for key, value in config_dict.items():
                api_configs_list.append(cm_client.ApiConfig(name=key, value=value))
            api_service_config = ApiServiceConfig(items=api_configs_list)

            api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
            api_response = api_instance.update_service_config(self.cluster_name, service_name, body=api_service_config)
            LOG.debug("Response %s" % api_response)
            return True, f"Service config for '{service_name}' successfully updated with key-values {config_dict}"
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->service_command_by_name: %s\n" % e)
            return False, str(e)

    def get_current_config_value(self, service_name: str, config_name: str):
        configs = self._read_service_config(service_name)
        for config in configs.items:
            if config.name == config_name:
                return config.value if config.value else config.default
        raise StopIteration

    def _read_service_config(self, service_name: str):
        LOG.info("Getting service configurations for '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.read_service_config(self.cluster_name, service_name, view='full')
            LOG.debug("Response %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->service_command_by_name: %s\n" % e)
            raise e

    def update_cm_config(self, config_dict: dict):
        LOG.info("Attempting to update CM configuration values...")
        config_list = []
        for key, value in config_dict.items():
            config_list.append(cm_client.ApiConfig(name=key, value=value))
        configs = cm_client.ApiConfigList(config_list)

        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.update_config(body=configs)
            LOG.debug("Response %s" % api_response)
            return True, "CM configurations has been updated..."
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->update_config: %s\n" % e)
            return False, str(e)

    def get_all_mgmt_role_config_groups(self) -> ApiRoleConfigGroupList:
        api_instance = cm_client.MgmtRoleConfigGroupsResourceApi(cm_client.ApiClient())
        return api_instance.read_role_config_groups()

    def update_mgmt_role_config(self, role_config_group_name: str, config_dict: dict):
        LOG.info(f"Attempting to update CM Management '{role_config_group_name}' Role's config...")
        config_list = []
        for key, value in config_dict.items():
            config_list.append(cm_client.ApiConfig(name=key, value=value))
        configs = cm_client.ApiConfigList(config_list)

        api_instance = cm_client.MgmtRoleConfigGroupsResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.update_config(role_config_group_name, body=configs)
            LOG.debug("Response %s" % api_response)
            return True, "CM configurations has been updated..."
        except ApiException as e:
            LOG.error("Exception when calling MgmtRoleConfigGroupsResourceApi->update_config: %s\n" % e)
            return False, str(e)

    def check_if_config_is_present_for_mgmt_role_config_group(self, role_config_group_name: str, config_name: str) \
            -> bool:
        api_instance = cm_client.MgmtRoleConfigGroupsResourceApi(cm_client.ApiClient())
        api_response: ApiConfigList = api_instance.read_config(role_config_group_name, view='full')
        return self._check_if_config_is_present(api_response, config_name)

    def _check_if_config_is_present(self, config_list: ApiConfigList, config_name: str):
        config_items: list[ApiConfig] = config_list.items
        try:
            next(conf for conf in config_items if conf.name == config_name)
            return True
        except StopIteration:
            return False

    def read_roles(self, service_name: str):
        LOG.info("Reading roles for service '%s'" % service_name)
        api_instance = cm_client.RolesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.read_roles(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling RolesResourceApi->read_roles: %s\n" % e)

    def run_role_command(self, command_name: str, service_name: str, role_names: list):
        LOG.info("Running '%s' role command for service '%s'" % (command_name, service_name))
        api_instance = cm_client.RoleCommandsResourceApi(cm_client.ApiClient())

        body = cm_client.ApiRoleNameList(role_names)
        try:
            api_response = api_instance.role_command_by_name(self.cluster_name, command_name, service_name, body=body)
            LOG.debug("Response %s" % api_response)
            for role_name in role_names:
                success, message = self.wait_for_role_commands_to_run(role_name, service_name)
                if not success:
                    return False, message
                else:
                    LOG.info(message)
            else:
                return True, f"Running role command {command_name} on service {service_name} successful"
        except ApiException as e:
            LOG.error("Exception when calling RolesResourceApi->read_roles: %s\n" % e)
            return False, str(e)

    def get_cluster(self, cluster_name: str = None):
        if cluster_name is None:
            cluster_name = self.cluster_name

        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        try:
            # Reads information about a cluster.
            return api_instance.read_cluster(cluster_name)
        except ApiException as e:
            LOG.error("Exception when calling ClustersResourceApi->read_cluster: %s\n" % e)
        return None

    def is_cluster_exist(self, cluster_name: str = None):
        return self.get_cluster(cluster_name) is not None

    def is_management_service_exists(self):
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())
        LOG.info(f"Checking whether Cloudera Management Service already exists on cluster {self.cluster_name}, on host"
                 f"{self.hostname}")
        try:
            api_instance.read_service(view='summary')
            return True
        except ApiException:
            return False

    def create_cloudera_management_service(self):
        """
        Attempt to create cloudera management service

        :return: Tuple of ( success , message )
        """
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())
        LOG.info(f"Attempting to create empty management service on cluster {self.cluster_name}, on host"
                 f"{self.hostname}")
        try:
            body = cm_client.ApiService(display_name='Cloudera Management Service')
            api_instance.setup_cms(body=body)
            return True, "Cloudera Management Service successfully created."
        except ApiException as e:
            return False, str(e)

    def update_role_config_group_property(self, service_name: str, role_config_group_display_name: str,
                                          configuration_name: str, configuration_value: str):
        """
        Attempt to update a property to a given value in a role configuration group of a service

        :return: Tuple of ( success , message )
        """
        try:
            role_config_group_name = self. \
                _get_role_config_group_name_from_all_role_config_groups(service_name, role_config_group_display_name)
            message = f"Updated {service_name} service's {role_config_group_display_name}  role configuration group's" \
                      f" {configuration_name} configuration to the following value: '{configuration_value}'"
            config_to_set = {configuration_name: configuration_value}
            self.update_role_config_group_configuration(service_name, role_config_group_name,
                                                        config_to_set, message)
            return True, message
        except ApiException as e:
            return False, str(e)
        except StopIteration:
            return False, "Couldn't find role config group name among role config groups"

    def get_all_services(self) -> list:
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        return api_instance.read_services(self.cluster_name).items

    def get_all_services_exception_handled_version(self):
        try:
            return True, self.get_all_services()
        except ApiException as e:
            return False, str(e)

    def get_role_config_groups_for_service(self, service_name: str) -> ApiRoleConfigGroupList:
        """
        Get all the role configs groups that are associated to a given service

        :param: service_name
        :return: ApiRoleConfigGroupList
        :except: ApiException
        """
        api_instance = cm_client.RoleConfigGroupsResourceApi(cm_client.ApiClient())
        return api_instance.read_role_config_groups(self.cluster_name, service_name)

    def check_if_config_is_present_for_role_config_group(
            self, service_name: str, role_config_group_name: str, config_name: str) -> bool:
        api_instance = cm_client.RoleConfigGroupsResourceApi(cm_client.ApiClient())
        api_response: ApiConfigList = api_instance.read_config(
            self.cluster_name, role_config_group_name, service_name, view='full')
        return self._check_if_config_is_present(api_response, config_name)

    def _get_role_config_group_name_of_role_config_display_name(self, role_config_groups: ApiRoleConfigGroupList,
                                                                display_name: str):
        """
        Find the real role config group name based on the display role config group name

        :param: role_config_groups: The list to search in
        :param: display_name: Filter criteria
        :return: Matching role_config_group.name
        :except: StopException
        """
        return next(role_config_group.name for role_config_group in role_config_groups.items
                    if role_config_group.display_name == display_name)

    def _get_role_config_group_name_from_all_role_config_groups(self, service_name: str, display_name: str):
        role_config_groups = self.get_role_config_groups_for_service(service_name)
        return self._get_role_config_group_name_of_role_config_display_name(role_config_groups,
                                                                            display_name)

    def update_role_config_group_configuration(self, service_name: str, role_config_group_name: str,
                                               config_dict: dict, message: str = ""):
        api_instance = cm_client.RoleConfigGroupsResourceApi(cm_client.ApiClient())
        config_list = []
        for key, value in config_dict.items():
            config_list.append(cm_client.ApiConfig(name=key, value=value))
        body = cm_client.ApiConfigList(config_list)

        api_instance.update_config(self.cluster_name, role_config_group_name,
                                   service_name, message=message, body=body)

    def read_configuration_from_role_config_group(self, service_name: str, role_config_group_display_name: str,
                                                  configuration_name: str):
        try:
            role_config_group_name = self._get_role_config_group_name_from_all_role_config_groups(service_name,
                                                                                                  role_config_group_display_name)
            api_instance = cm_client.RoleConfigGroupsResourceApi(cm_client.ApiClient())
            api_response = api_instance.read_config(self.cluster_name, role_config_group_name,
                                                    service_name, view='summary')
            return True, next(config.value for config in api_response.items if config.name == configuration_name)
        except ApiException as e:
            return False, str(e)
        except StopIteration:
            return False, "Couldn't find role config group name among role config groups"

    def is_tls_enabled(self):
        """
        API call to CM server whether TLS is enabled.

        :return:  bool
        :except ApiException
        """
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        return api_instance.is_tls_enabled(self.cluster_name)

    def get_hostnames_belonging_to_service(self, service_name: str):
        try:
            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_response = api_instance.export(self.cluster_name, export_auto_config=False)
            # --------------
            # we filter by 'service_name' which is a substring of a role name
            # the list of role names is: host_template.role_config_groups_ref_names
            # and these values are represented like this:
            #     <service name>-<functionality>-<some number>
            # Example: zookeeper-SERVER-1
            # --------------
            # we need to collect 'host_template.ref_name' values

            host_template_ref_names = list(host_template.ref_name for host_template in api_response.host_templates
                                           if any(service_name in role_name
                                                  for role_name in host_template.role_config_groups_ref_names))

            # for testing:
            #host_template_ref_names = [
            #     'HostTemplate-3-from-ctr-e172-[004-012][0-3].hwx.site',
            #     'HostTemplate-3-from-ctr-e172-[0, 2-4, 6]-1620330694487-708631-[9, 11]-01-[000010-000011, 000013].hwx.site',
            #     'HostTemplate-2-from-ctr-e172-1620330694487-708631-01-000005.hwx.site']
            hostnames_belonging_to_service = []
            for host_template_ref_name in host_template_ref_names:
                # if a host_template_ref_name contains a '[' it is treated as a special kind of regular expression
                if "[" in host_template_ref_name:
                    hostnames_belonging_to_service.extend(self.get_hostnames_from_regex(host_template_ref_name))
                else:
                    # not a 'regex', just append to the result list
                    hostnames_belonging_to_service.append(host_template_ref_name)

            # host_template.ref_name values are represented like this:
            #     HostTemplate-<some number>-from-<hostname>
            # Example: HostTemplate-0-from-serverhostname.org
            # so this "HostTemplate-\d+-from-" should be cut off
            hostnames_belonging_to_service = list(map(lambda a: re.search(r"HostTemplate-\d+-from-(.*)", a).group(1),
                                                      hostnames_belonging_to_service))
            return True, json.dumps(hostnames_belonging_to_service)
        except ApiException as e:
            return False, str(e)

    def get_hostnames_from_regex(self, host_template_ref_name: str):
        # Example input:
        # 'HostTemplate-3-from-ctr-e172-[0, 2-4, 6]-1620330694487-708631-[9, 11]-01-[000010-000011, 000013].hwx.site'
        opening_bracket_splits = host_template_ref_name.split("[")
        closing_bracket_splits = list(i.split("]") for i in opening_bracket_splits)
        # after closing bracket split:
        # ['HostTemplate-3-from-ctr-e172-', '0, 2-4, 6', '-1620330694487-708631-',
        # '9, 11', '-01-', '000010-000011, 000013', '.hwx.site']
        flat_list = reduce(operator.concat, closing_bracket_splits)
        comma_splits = list(i.split(", ") for i in flat_list)
        # at this point comma_splits will contain:
        # list w/ 1 element => these will remain static
        # list w/ multiple elements:
        #     'singular elements' => these will be used as they appear in the cartesian product at the end
        #     'range elements' => these have to be converted into ranges and then back again into list elements
        # Example:
        # comma_splits = [
        # ['HostTemplate-3-from-ctr-e172-'],
        # ['0', '2-4', '6'], # here '0' and '6' are singular, '2-4' is a range
        # ['-1620330694487-708631-'],
        # ['9', '11'], # two singular elements
        # ['-01-'],
        # ['000010-000011', '000013'], # '000010-000011' is also range
        # ['.hwx.site']
        # ]
        # a mapping will handle the range elements and create list elements from those as well
        # a flattening operation is needed after the mapping
        # to merge the 'singular' elements with the expanded 'range' elements
        # Example after map & flat:
        # ranges_expanded =[
        # ['HostTemplate-3-from-ctr-e172-'],
        # ['0', '2', '3', '4', '6'],
        # ['-1620330694487-708631-'],
        # ['9', '11'],
        # ['-01-'],
        # ['000010', '000011', '000013'],
        # ['.hwx.site']
        # ]
        # This is exactly what we need as input for a cartesian product.
        # That is, tuples (which are joined into strings at the end) of every possible
        # combination of these lists. => itertools.product(*ranges_expanded)
        # end result:
        # [
        # 'HostTemplate-3-from-ctr-e172-0-1620330694487-708631-9-01-000010.hwx.site',
        # 'HostTemplate-3-from-ctr-e172-0-1620330694487-708631-9-01-000011.hwx.site',
        # 'HostTemplate-3-from-ctr-e172-0-1620330694487-708631-9-01-000013.hwx.site',
        # ...
        # ]

        ranges_expanded = list(
            reduce(operator.concat, map(self.range_element_to_string_list, a)) for a in comma_splits)

        return map(lambda a: "".join(a), itertools.product(*ranges_expanded))

    def range_element_to_string_list(self, elem):
        if re.match("^[0-9]+-[0-9]+$", elem):
            elem_range = elem.split("-")
            range_start = elem_range[0]
            range_end = elem_range[1]
            range_format_len = len(range_start)
            token_range = range(int(range_start), int(range_end) + 1)
            # padding character is '0'
            result = list(str(i).rjust(range_format_len, '0') for i in token_range)
        else:
            result = [elem]
        return result

    def attempt_to_get_configuration_object_from_role_configuration_groups(self,
                                                                           service_object: ApiClusterTemplateService,
                                                                           configuration_name: str):
        # data structure:
        #   service_object.role_config_groups is a list of role_config_groups, where each element is =>
        #     a list of ApiClusterTemplateRoleConfigGroup elements, where each element has =>
        #       a dictionary key named 'configs' that has a list as a value which contains =>
        #         ApiClusterTemplateConfig elements, where its 'name' fields are the one used for filtering
        return next(api_cluster_template_config
                    for api_cluster_template_role_config_group in service_object.role_config_groups
                    if api_cluster_template_role_config_group.configs
                    for api_cluster_template_config in api_cluster_template_role_config_group.configs
                    if api_cluster_template_config.name == configuration_name)

    def get_configuration_value_from_configuration_object(self,
                                                          configuration_object: ApiClusterTemplateConfig,
                                                          all_variables: list):
        configuration_value = None
        if configuration_object.value is not None:
            configuration_value = configuration_object.value
        elif configuration_object.variable is not None:
            configuration_value = next(service_variable.value for service_variable in all_variables
                                       if service_variable.name == configuration_object.variable)
        return configuration_value

    def query_configuration_value_from_role_configuration_groups(self, service_name: str,
                                                                 configuration_name: str):
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.export(self.cluster_name, export_auto_config=False)
            service_object = None
            configuration_object = None
            try:
                service_object = next(
                    api_cluster_template_service for api_cluster_template_service in api_response.services
                    if api_cluster_template_service.ref_name == service_name)
                configuration_object = self.attempt_to_get_configuration_object_from_role_configuration_groups(
                    service_object,
                    configuration_name)
                configuration_value = self.get_configuration_value_from_configuration_object(configuration_object,
                                                                                             api_response.instantiator.variables)
            except StopIteration:
                msg = None
                if service_object is None:
                    msg = self.generate_service_name_not_found_message(service_name, api_response.services)
                elif configuration_object is None:
                    msg = self.generate_configuration_name_not_found_message(service_name, configuration_name,
                                                                             service_object.role_config_groups)
                return False, msg
            return True, configuration_value
        except ApiException as e:
            return False, str(e)

    def generate_service_name_not_found_message(self, service_name: str, available_services: list):
        return f"Service name '{service_name}' is not found in API response, " \
               f"please choose ONE from the following available values:" \
               + ",".join(list(api_cluster_template_service.ref_name for api_cluster_template_service in
                               available_services))

    def generate_configuration_name_not_found_message(self, service_name: str,
                                                      configuration_name: str,
                                                      role_configuration_groups: list):
        return f"Configuration name '{configuration_name}' for service name '{service_name}' " \
               f"is not found in API response, " \
               f"please choose ONE from the following available values: " \
               + ", ".join(list(config.name for api_cluster_template_config in role_configuration_groups
                                if api_cluster_template_config.configs
                                for config in api_cluster_template_config.configs))

    def import_kerberos_admin_credentials(self, username: str, password: str):
        LOG.info("Attempting to import kerberos admin credentials to CM...")
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.import_admin_credentials(username=username, password=password)
            LOG.debug(f"Response: {api_response}")
            return True, "Imported kerberos admin credentials to CM"
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->import_admin_credentials: %s\n" % e)
            return False, str(e)

    def configure_cluster_for_kerberos(self):
        LOG.info("Calling configure for kerberos API endpoint on CM...")
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())

        body = self.create_configure_kerberos_argument_body()

        try:
            api_response = api_instance.configure_for_kerberos(self.cluster_name, body=body)
            LOG.debug(f"Response: {api_response}")
            return True, "Configure for kerberos API call successful"
        except ApiException as e:
            LOG.error("Exception when calling ClustersResourceApi->configure_for_kerberos: %s\n" % e)
            return False, str(e)

    def create_configure_kerberos_argument_body(self):
        success_transceiver_port, response_transceiver_port = \
            self.read_configuration_from_role_config_group("hdfs", "hdfs-DATANODE", "dfs_datanode_port");
        success_datanode_web_port, response_datanode_web_port = \
            self.read_configuration_from_role_config_group("hdfs", "hdfs-DATANODE", "dfs_datanode_http_port");
        # since ApiConfigureForKerberosArguments doesn't correctly fill out default values,
        # we'll have to do it manually
        datanode_transceiver_port = int(response_transceiver_port) if success_transceiver_port else 1004
        datanode_web_port = int(response_datanode_web_port) if success_datanode_web_port else 1006
        return cm_client.ApiConfigureForKerberosArguments(datanode_transceiver_port=datanode_transceiver_port,
                                                          datanode_web_port=datanode_web_port)

    def delete_kerberos_credentials(self, delete_credentials_mode='all'):
        LOG.info(f"Calling Kerberos credential deletion with mode: {delete_credentials_mode}")
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.delete_credentials_command(delete_credentials_mode=delete_credentials_mode)
            LOG.debug(f"Response: {api_response}")
            return True, "Kerberos credentials has been deleted"
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->delete_credentials_command: %s\n" % e)
            return False, str(e)

    def generate_missing_kerberos_credentials(self):
        LOG.info("Initiation of Kerberos credential generation...")
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.generate_credentials_command()
            LOG.debug(f"Response: {api_response}")
            return True, "Kerberos credential generation initiated..."
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->generate_credentials_command: %s\n" % e)
            return False, str(e)

    def read_cm_management_service(self) -> ApiService:
        LOG.info("Read CM Management Service")
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())

        try:
            return api_instance.read_service()
        except ApiException as e:
            LOG.error("Exception when calling MgmtServiceResourceApi->read_service: %s\n" % e)
            raise e

    def stop_cm_management_service(self, timeout_sec: int = 360):
        LOG.info("Initiating STOP command for CM Management Service")
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.stop_command()
            LOG.debug(f"Response: {api_response}")
            expected_service_state = {'service_state': 'STOPPED'}
            return self.wait_for_service_state('mgmt', expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling MgmtServiceResourceApi->stop_command: %s\n" % e)
            return False, str(e)

    def start_cm_management_service(self, timeout_sec: int = 360):
        LOG.info("Initiating START command for CM Management Service")
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.start_command()
            LOG.debug(f"Response: {api_response}")
            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            return self.wait_for_service_state('mgmt', expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling MgmtServiceResourceApi->start_command: %s\n" % e)
            return False, str(e)

    def restart_cm_management_service(self, timeout_sec: int = 360):
        LOG.info("Initiating RESTART command for CM Management Service")
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.restart_command()
            LOG.debug(f"Response: {api_response}")
            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            return self.wait_for_service_state('mgmt', expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling MgmtServiceResourceApi->restart_command: %s\n" % e)
            return False, str(e)

    def enable_tls_for_services(self):
        try:
            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_instance.configure_auto_tls_services_command(self.cluster_name)
            return True, f"Successfully enabled auto TLS for services in cluster '{self.cluster_name}'!"
        except ApiException as e:
            return False, str(e)

    def enable_tls(self, remote_username: str, **kwargs):
        """
        Enabling TLS in CM. No Root CA or custom cert is given, CM generates everything in this case.

        :param remote_username: The username to pass to CM server so that it can deploy the certificates to agents
        :param kwargs: can be 'private_key_file' or 'password'
        :return: tuple of success, message
        """
        try:
            api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())
            if kwargs.get('private_key_file'):
                text_file = open(kwargs.get('private_key_file'))
                private_key = text_file.read()
                text_file.close()
                body = cm_client.ApiGenerateCmcaArguments(user_name=remote_username, private_key=private_key)
            elif 'password' in kwargs:
                body = cm_client.ApiGenerateCmcaArguments(user_name=remote_username, password=kwargs.get('password'))
            else:
                return False, "No private_key_file or password parameter. Please supply one or the other. If both" \
                              "is supplied, private_key_file takes precedence."
            api_instance.generate_cmca(body=body)

            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_instance.configure_auto_tls_services_command(self.cluster_name)
            return True, f"Successfully enabled auto TLS for services in cluster '{self.cluster_name}'!"
        except ApiException as e:
            return False, str(e)

    def start_all_services(self):
        try:
            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_response = api_instance.start_command(self.cluster_name)

            if api_response.success is not None and not api_response.success:
                return api_response.active, api_response.result_message

            return self.wait_until_all_services_are_started()
        except ApiException as e:
            return False, str(e)

    def wait_until_all_services_are_started(self):
        try:
            all_services = self.get_all_services()
        except ApiException as e:
            return False, f"Couldn't read services from API {str(e)}"
        expected_service_state = {'service_state': 'STARTED', 'health_summary': '!DISABLED'}
        all_services = list(service for service in all_services if service.service_state != 'NA')
        for service in all_services:
            is_service_started, message = self.wait_for_service_state(service.name, expected_service_state)
            if not is_service_started:
                return is_service_started, message
        return True, "All services have been successfully started"

    def deploy_cluster_client_config(self):
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        body = cm_client.ApiHostRefList(items=[])
        try:
            api_response = api_instance.deploy_cluster_client_config(self.cluster_name, body=body)
            if api_response.active:
                return True, "Successfully launched Kerberos configuration deployment to all hosts!"
            else:
                return False, "Could not launch Kerberos configuration deployment to hosts!"
        except ApiException as e:
            return False, str(e)
